open ExtList

module Build (Env : Env.EnvSig) (LJT : LJT.LJTSig) = struct
open LJT

let problems : (problem, string) Hashtbl.t = Hashtbl.create 37
let n_problems = ref 0

let problems_labels = ref ([] : (string * string) list)

let psfrag_text = ref ([] : (string * string) list)
let psfrag_text_cont = ref ([] : (string * string) list)

let id_psfrag = ref 0
let id_psfrag_cont = ref 0

let output_psfrag () =
  let ct = [ "Ax", "\\Ax";
             "ImpL", "\\ImpL";
             "ImpR", "\\ImpR";
             "AndL1", "\\AndLp{1}"; (* TODO *)
             "AndL2", "\\AndLp{2}";
             "AndR", "\\AndR";
             "OrL", "\\OrL";
             "OrR1", "\\OrRp{1}";
             "OrR2", "\\OrRp{2}"
           ] in
  let psfrag = open_out "psfrag.tex" in
    List.iter (fun (lbl, txt) -> 
                 Printf.fprintf psfrag "\\psfrag{%s}{$%s$}\n" lbl txt) 
      (ct @ !psfrag_text_cont);
    List.iter (fun (lbl, txt) -> 
                 Printf.fprintf psfrag "\\psfrag{%s}{$\\!\\!%s$}\n" lbl txt) 
      !psfrag_text

let generate_string_id_aux id c len = 
  let i = !id in
    incr id;
    let si = string_of_int i in
      String.make (len - String.length si) c ^ si

let generate_string_id = generate_string_id_aux id_psfrag 'a'
let generate_string_id_cont = generate_string_id_aux id_psfrag_cont 'b'

let get_size s = 
  let fdout = open_out "text.tex" in
    Printf.fprintf fdout "%s" s;
    close_out fdout;
    let fd = Unix.open_process_in "latex size.tex | grep SIZE" in
    let scan = Scanf.Scanning.from_channel fd in
    let rfloat = Scanf.bscanf scan "SIZE1: %fpt\nSIZE2: %fpt\n" 
                   (fun s1 s2 -> s2 /. s1) in
      close_in fd;
      int_of_float (rfloat /. 1.2)

let print_problems_labels () = 
  List.iter (fun (n, label) ->
               if !Options.result = Options.LaTeX then
                 let size = get_size label in
                 let id = generate_string_id size in
                   Printf.printf "%s [shape=box, color=black, label=\"%s\"]\n" 
                     n id;
                   psfrag_text := (id, label) :: !psfrag_text
               else
                 Printf.printf "%s [shape=box, color=grey40, label=\"%s\"]\n"
                   n label)
    !problems_labels

let n_rules = ref 0

let rec node_name_of_problem ?initial_env pb =
  try Hashtbl.find problems pb
  with Not_found ->
    let r = Printf.sprintf "pb%d" !n_problems in
      Hashtbl.add problems pb r;
      incr n_problems;
      let n_proofs =
        try 
          Printf.sprintf ": %s" (string_of_proofs_found 
                                   (retrieve_problem_node pb).nb_proofs)
        with Not_found -> "" in
      let tpb = string_of_problem ?initial_env pb ^ n_proofs in
         problems_labels := (r, match !Options.result with
                               | Options.Text -> String.escaped tpb
                               | Options.LaTeX -> tpb) :: !problems_labels;
        r

let rec output_problem ?initial_env pb =
  if Hashtbl.mem problems pb = false then
    let node_name = node_name_of_problem ?initial_env pb in
      try 
        let node = retrieve_problem_node pb in
          if node.hypotheses <> [] then
            let node_rules = name_rules node_name node.hypotheses in
              Printf.printf "%s -> {%s} [color=grey40, style=dotted]\n" node_name 
                (String.concat "; " (List.map fst node_rules));
              output_rules ?initial_env node_name node_rules;
      with Not_found -> ()
and output_rules ?initial_env pb rules = 
  let rec aux = function
    | name, Ax -> 
        Printf.printf "%s[shape=box, style=dotted, label=\"Ax\"]" name
    | name, Cont (prop, p) ->
        output_problem ?initial_env p;
        if !Options.result = Options.LaTeX then
          let s = Printf.sprintf "\\Cont(%s)" (Prop.to_string prop) in
          let len = get_size s in
          let id = generate_string_id_cont len in
            Printf.printf "%s [shape=box,style=dotted, label=\"%s\"]\n" name id;
            Printf.printf "%s -> %s\n" name (node_name_of_problem p);
            psfrag_text_cont := (id, s) :: !psfrag_text_cont
        else (
          Printf.printf "%s [shape=box,style=dotted, label=\"Cont %s\"]\n" name
            (String.escaped (Prop.to_string_text prop));
          Printf.printf "%s -> %s\n" name (node_name_of_problem p))
    | name, ImpR p ->
        output_problem ?initial_env p;
        Printf.printf "%s [shape=box, style=dotted, label=\"ImpR\"]\n" name;
        Printf.printf "%s -> %s\n" name (node_name_of_problem p)
    | name,  AndL ((i, _), p) ->
        output_problem ?initial_env p;
        Printf.printf "%s [shape=box, style=dotted, label=\"AndL%d\"]\n" 
          name (i+1);
        Printf.printf "%s -> %s\n" name (node_name_of_problem p)
    | name, ImpL (p1, p2) ->
        output_problem ?initial_env p1;
        output_problem ?initial_env p2;
        Printf.printf "%s [shape=box, style=dotted, label=\"ImpL\"]\n" name;
        Printf.printf "%s ->%s[label=\"1\"]\n" name (node_name_of_problem p1);
        Printf.printf "%s ->%s[label=\"2\"]\n" name (node_name_of_problem p2)
    | name, AndR l ->
        List.iter (output_problem ?initial_env) l;
        Printf.printf "%s [shape=box, style=dotted, label=\"AndR\"]\n" name;
        List.iteri (fun i p -> Printf.printf "%s -> %s [label=\"%d\"]\n" 
                      name (node_name_of_problem p) i) l
    | name, OrL (_, l) ->
        List.iter (output_problem ?initial_env) l;
        Printf.printf "%s [shape=box, style=dotted, label=\"OrL\"]\n" name;
        List.iteri (fun i p -> Printf.printf "%s -> %s [label=\"%d\"]\n"
                      name (node_name_of_problem p) i) l
    | name,  OrR (_, (i, _), p) ->
        output_problem ?initial_env p;
        Printf.printf "%s [shape=box, style=dotted, label=\"OrR%d\"]\n" 
          name (i+1);
        Printf.printf "%s -> %s\n" name (node_name_of_problem p)
  in 
    List.iter aux rules
and name_rules pb = List.mapi (fun i r -> (Printf.sprintf "%sr%d" pb i), r)

end
