open ExtList

module LJTSet = LJT.Build(Env.EnvSet)
module DotSet = Dot.Build(Env.EnvSet)(LJTSet)

module LJTMSet = LJT.Build(Env.EnvMultiSet)
module DotMSet = Dot.Build(Env.EnvMultiSet)(LJTMSet)
module WalkMSet = Walk.Build(LJTMSet)
module InterpretProof = InterpretProof.Build(LJTMSet)

module Strategy = LJT.Strategy

(* TODO : it should be possible to cast a dictionnary into a set *)
let oracle d (env, stoup, goal) = 
  let pb = { LJTSet.env = Env.EnvSet.of_list env ; 
             LJTSet.stoup = stoup ; 
             LJTSet.goal = goal } in 
  let n, c = LJTSet.solve_problem { Strategy.counting with 
                                      Strategy.max_search_depth = d } pb in
    if d = None then assert c;
    Some n

let choose_strategy () = match !Options.explore_type with
  | Options.Counting -> LJT.Strategy.counting
  | Options.TermEnumeration -> LJT.Strategy.explore_provable
  | Options.ExploreAll -> LJT.Strategy.explore_all

let update_depth_search s = 
  { s with LJT.Strategy.max_search_depth = !Options.depth_search }

let update_oracle s = 
  if !Options.use_oracle then
    { s with Strategy.oracle = oracle !Options.depth_oracle }
  else s

module IntSet = Set.Make(struct type t = int let compare = (-) end)
let rec find_types p = match Prop.prop_of_t p with
  | Prop.Var i -> IntSet.singleton i
  | Prop.Arrow (p1, p2) -> IntSet.union (find_types p1) (find_types p1)
  | Prop.And l | Prop.Or (_, l) -> List.fold_left 
      (fun s p -> IntSet.union s (find_types p)) IntSet.empty l

let print_types pb =
  IntSet.iter (fun i -> Printf.printf "  type t%d\n" i)
    (List.fold_left (fun s p -> IntSet.union s (find_types p))
       (find_types pb.LJTMSet.goal) (Env.EnvMultiSet.elems_set pb.LJTMSet.env))

let print_types_parseably pb =
  IntSet.iter (fun i -> Printf.printf "T %d\n" i)
    (List.fold_left (fun s p -> IntSet.union s (find_types p))
       (find_types pb.LJTMSet.goal) (Env.EnvMultiSet.elems_set pb.LJTMSet.env))

let set_graph_style () =
  Options.result := if !Options.output = Options.GraphText then 
    Options.Text else Options.LaTeX

let initial_env_name () =
  match !Options.result with
    | Options.Text -> "E"
    | Options.LaTeX -> "\\Sigma"

let main_set pb =
  let strategy = update_depth_search (choose_strategy ()) in
  let sols = LJTSet.solve_problem strategy pb in
  (* reexplore false; *)
    match !Options.output with
      | Options.NbProofs -> 
          let hpb = LJTSet.retrieve_problem_node pb in
            Printf.printf "%s solutions\n"
              (LJTSet.string_of_proofs_found hpb.LJTSet.nb_proofs)
      | Options.GraphText | Options.GraphLaTeX ->
          set_graph_style ();
          print_string "digraph G {\n";
          DotSet.output_problem ~initial_env:(pb.LJTSet.env, 
                                              initial_env_name ()) pb;
          DotSet.print_problems_labels ();
          DotSet.output_psfrag ();
          print_string "}\n"
      | Options.CamlCode ->
          prerr_string "Producing lambda-terms using LJT^Set is impossible\n";
          exit 1

let main_mset pb = 
  let strategy = update_oracle (update_depth_search (choose_strategy ())) in
  let sols = LJTMSet.solve_problem strategy pb in
  (* reexplore false; *)
    match !Options.output with
      | Options.NbProofs -> 
          let hpb = LJTMSet.retrieve_problem_node pb in
            Printf.printf "%s solutions\n"
              (LJTMSet.string_of_proofs_found hpb.LJTMSet.nb_proofs)
      | Options.GraphText | Options.GraphLaTeX ->
          set_graph_style ();
          print_string "digraph G {\n";
          DotMSet.output_problem ~initial_env:(pb.LJTMSet.env,
                                              initial_env_name ()) pb;
          DotMSet.print_problems_labels ();
          DotMSet.output_psfrag ();
          print_string "}\n"
      | Options.CamlCode ->
          let reff = ref 0 in
          let l = WalkMSet.walk_problem !Options.walk_depth pb in
          let env = EnvLambda.initial_env 
                      (Env.EnvMultiSet.elems_multiset pb.LJTMSet.env) in
          let v = !LambdaBar.ref_var in
            print_string "module Gen (Env: sig\n";
            print_types pb;
            EnvLambda.print_env env;
            print_string "end) =\nstruct\nopen Env\n";
            List.iter (fun t -> 
                         Printf.printf "let f%d : %s= %s;;\n" 
                         !reff (Prop.to_string_ocaml pb.LJTMSet.goal)
                         (LambdaBar.string_of_lterm (LambdaBar.lterm_of_t t));
                         incr reff)
              (List.map (fun p -> LambdaBar.reset_var v;
                           InterpretProof.interpret_proof pb env p) l);
            print_string "end\n"
      | Options.Parseable ->
          let reff = ref 0 in
          let l = WalkMSet.walk_problem !Options.walk_depth pb in
          let env = EnvLambda.initial_env 
                      (Env.EnvMultiSet.elems_multiset pb.LJTMSet.env) in
          let v = !LambdaBar.ref_var in
            print_types_parseably pb;
            EnvLambda.print_env_parseable env;
            List.iter (fun t -> 
                         Printf.printf "(F %d %s %s)\n" 
                         !reff (Prop.to_string_parseable pb.LJTMSet.goal)
                         (LambdaBar.string_of_lterm_parseable (LambdaBar.lterm_of_t t));
                         incr reff)
              (List.map (fun p -> LambdaBar.reset_var v;
                           InterpretProof.interpret_proof pb env p) l);
