(* camlp4r *)
type reg = State.reg
type t =
    ConditionalMove of reg * reg * reg
  | ArrayIndex of reg * reg * reg
  | ArrayAmendment of reg * reg * reg
  | Addition of reg * reg * reg
  | Multiplication of reg * reg * reg
  | Division of reg * reg * reg
  | NotAnd of reg * reg * reg
  | Halt
  | Allocation of reg * reg
  | Abandonment of reg
  | Output of reg
  | Input of reg
  | LoadProgram of reg * reg
  | Orthography of reg * int32

(* value zero = 0b00000000000000000000000000000000; *)
let op_mask = 0b11110000000000000000000000000000l, 32 - 4
let a_mask = 0b00000000000000000000000111000000l, 6
let b_mask = 0b00000000000000000000000000111000l, 3
let c_mask = 0b00000000000000000000000000000111l, 0
let a2_mask = 0b00001110000000000000000000000000l, 32 - 4 - 3
let v_mask = 0b00000001111111111111111111111111l, 0

let m_int32 (mask, shift) x =
  Int32.shift_right_logical (Int32.logand mask x) shift

let m mask x = Int32.to_int (m_int32 mask x)

let of_int32 x =
  let opcode = m op_mask x in
  match opcode with
    13 ->
      let a = m a2_mask x
      and v = m_int32 v_mask x in
      Orthography (a, v)
  | _ ->
      let a = m a_mask x in
      let b = m b_mask x in
      let c = m c_mask x in
      match opcode with
        0 -> ConditionalMove (a, b, c)
      | 1 -> ArrayIndex (a, b, c)
      | 2 -> ArrayAmendment (a, b, c)
      | 3 -> Addition (a, b, c)
      | 4 -> Multiplication (a, b, c)
      | 5 -> Division (a, b, c)
      | 6 -> NotAnd (a, b, c)
      | 7 -> Halt
      | 8 -> Allocation (b, c)
      | 9 -> Abandonment c
      | 10 -> Output c
      | 11 -> Input c
      | 12 -> LoadProgram (b, c)
      | x ->
          invalid_arg
            ("Instruction.of_int32: unknown opcode " ^ string_of_int x)
let to_int32 _ = 0l

let pp = Format.fprintf
let r f = pp f "%i"
let print f =
  function
    ConditionalMove (a, b, c) ->
      pp f "ConditionalMove a:%a b:%a c:%a" r a r b r c
  | ArrayIndex (a, b, c) -> pp f "ArrayIndex a:%a b:%a c:%a" r a r b r c
  | ArrayAmendment (a, b, c) ->
      pp f "ArrayAmendment a:%a b:%a c:%a" r a r b r c
  | Addition (a, b, c) -> pp f "Addition a:%a b:%a c:%a" r a r b r c
  | Multiplication (a, b, c) ->
      pp f "Multiplication a:%a b:%a c:%a" r a r b r c
  | Division (a, b, c) -> pp f "Division a:%a b:%a c:%a" r a r b r c
  | NotAnd (a, b, c) -> pp f "NotAnd a:%a b:%a c:%a" r a r b r c
  | Halt -> pp f "Halt"
  | Allocation (b, c) -> pp f "Allocation b:%a c:%a" r b r c
  | Abandonment c -> pp f "Abandonment c:%a" r c
  | Output c -> pp f "Output c:%a" r c
  | Input c -> pp f "Input c:%a" r c
  | LoadProgram (b, c) -> pp f "LoadProgram b:%a c:%a" r b r c
  | Orthography (a, v) -> pp f "Orthography a:%a v:%lu" r a v

let pretty_print f =
  function
    ConditionalMove (a, b, c) -> pp f "r%a := r%a unless r%a = 0." r a r b r c
  | ArrayIndex (a, b, c) -> pp f "r%a := r%a[r%a]." r a r b r c
  | ArrayAmendment (a, b, c) -> pp f "r%a[r%a] := r%a." r a r b r c
  | Addition (a, b, c) -> pp f "r%a := r%a + r%a." r a r b r c
  | Multiplication (a, b, c) -> pp f "r%a := r%a * r%a." r a r b r c
  | Division (a, b, c) -> pp f "r%a := r%a / r%a." r a r b r c
  | NotAnd (a, b, c) -> pp f "r%a := r%a nand r%a." r a r b r c
  | Halt -> pp f "halt."
  | Allocation (b, c) -> pp f "r%a := new array r%a." r b r c
  | Abandonment c -> pp f "free r%a." r c
  | Output c -> pp f "write r%a." r c
  | Input c -> pp f "read %a." r c
  | LoadProgram (b, c) -> pp f "load_program r%a.@ jump r%a." r b r c
  | Orthography (a, v) -> pp f "r%a := %lu" r a v

let to_string x =
  let b = Buffer.create 501 in
  let () = Format.bprintf b "%a@?" print x in Buffer.contents b
